package geniusweb.exampleparties.timedependentparty;

import java.io.IOException;
import java.math.BigDecimal;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Random;
import java.util.logging.Level;

import geniusweb.actions.Accept;
import geniusweb.actions.Action;
import geniusweb.actions.Offer;
import geniusweb.actions.PartyId;
import geniusweb.bidspace.AllBidsList;
import geniusweb.bidspace.BidsWithUtility;
import geniusweb.issuevalue.Bid;
import geniusweb.party.Capabilities;
import geniusweb.party.DefaultParty;
import geniusweb.party.inform.ActionDone;
import geniusweb.party.inform.Finished;
import geniusweb.party.inform.Inform;
import geniusweb.party.inform.Settings;
import geniusweb.party.inform.YourTurn;
import geniusweb.profile.utilityspace.LinearAdditiveUtilitySpace;
import geniusweb.profile.utilityspace.UtilitySpace;
import geniusweb.profileconnection.ProfileConnectionFactory;
import geniusweb.profileconnection.ProfileInterface;
import geniusweb.progress.Progress;
import geniusweb.progress.ProgressRounds;
import geniusweb.references.ProtocolRef;
import tudelft.utilities.logging.Reporter;

/**
 * General time dependent party. This is a simplistic implementation that does
 * brute-force search through the bidspace and can handle bidspace sizes up to
 * 2^31 (approx 1 billion bids). It may take excessive time and run out of time
 * on bidspaces > 10000 bids. In special cases it may even run out of memory,
 */
public abstract class TimeDependentParty extends DefaultParty {

	private ProfileInterface profileint;
	private PartyId me;
	private Progress progress;
	private Double minUtil = null, maxUtil = null;
	private Bid lastReceivedBid = null;
	private double tolerance;

	public TimeDependentParty() {
		super();
	}

	public TimeDependentParty(Reporter reporter) {
		super(reporter); // for debugging
	}

	@Override
	public Capabilities getCapabilities() {
		try {
			return new Capabilities(new HashSet<>(
					Arrays.asList(new ProtocolRef(new URI("SAOP")))));
		} catch (URISyntaxException e) {
			getReporter().log(Level.SEVERE, "Failed to create capabilities URI",
					e);
			return null;
		}
	}

	@Override
	public void notifyChange(Inform info) {
		try {
			if (info instanceof Settings) {
				Settings settings = (Settings) info;
				this.profileint = ProfileConnectionFactory
						.create(settings.getProfile().getURI(), getReporter());
				this.me = settings.getID();
				this.progress = settings.getProgress();
				computeMinMax();
			} else if (info instanceof ActionDone) {
				Action otheract = ((ActionDone) info).getAction();
				if (otheract instanceof Offer) {
					lastReceivedBid = ((Offer) otheract).getBid();
				}
			} else if (info instanceof YourTurn) {
				myTurn();
				if (progress instanceof ProgressRounds) {
					progress = ((ProgressRounds) progress).advance();
				}
			} else if (info instanceof Finished) {
				getReporter().log(Level.INFO, "Final ourcome:" + info);
			}
		} catch (Exception e) {
			getReporter().log(Level.SEVERE, "Failed to handle info", e);
		}
	}

	/**
	 * @return the E value that controls the party's behaviour. Depending on the
	 *         value of e, extreme sets show clearly different patterns of
	 *         behaviour [1]:
	 * 
	 *         1. Boulware: For this strategy e < 1 and the initial offer is
	 *         maintained till time is almost exhausted, when the agent concedes
	 *         up to its reservation value.
	 * 
	 *         2. Conceder: For this strategy e > 1 and the agent goes to its
	 *         reservation value very quickly.
	 * 
	 *         3. When e = 1, the price is increased linearly.
	 * 
	 *         4. When e = 0, the agent plays hardball.
	 */
	public abstract double getE();

	/******************* private support funcs ************************/

	private void myTurn() throws IOException {
		Action myAction = null;
		Bid bid = makeBid();
		UtilitySpace utilspace = getUtilSpace();

		if (lastReceivedBid != null && utilspace.getUtility(lastReceivedBid)
				.compareTo(utilspace.getUtility(bid)) > 0) {
			myAction = new Accept(me, lastReceivedBid);
		} else {
			myAction = new Offer(me, makeBid());
		}
		getConnection().send(myAction);

	}

	private UtilitySpace getUtilSpace() {
		return (LinearAdditiveUtilitySpace) profileint.getProfile();
	}

	/**
	 * Computes the fields minutil and maxUtil if not yet computed. minutil is
	 * the minimum utility of any bid, but at least the utility of the
	 * reservation bid. maxUtil is the maximum util of any bid.
	 * <p>
	 * TODO this is simplistic, very expensive method and may cause us to run
	 * out of time on large domains.
	 * <p>
	 * Can be called only after profileint has been set.
	 */
	private void computeMinMax() {
		if (this.maxUtil != null && this.minUtil != null)
			return;
		this.minUtil = 1d;
		this.maxUtil = 0d;
		UtilitySpace utilspace = getUtilSpace();
		AllBidsList allbidslist = new AllBidsList(
				profileint.getProfile().getDomain());
		for (Bid bid : allbidslist) {
			double util = utilspace.getUtility(bid).doubleValue();
			if (util < minUtil)
				minUtil = util;
			if (util > maxUtil)
				maxUtil = util;
		}
		double rv = utilspace.getUtility(utilspace.getReservationBid())
				.doubleValue();
		if (rv > minUtil)
			minUtil = rv;
		tolerance = 100 / allbidslist.size().longValue();
		tolerance = Math.max(tolerance, 0.1);
	}

	private Bid makeBid() {
		double time = progress.get(System.currentTimeMillis());
		double utilityGoal = utilityGoal(time, getE());
		BidsWithUtility options = new BidsWithUtility(
				(LinearAdditiveUtilitySpace) profileint.getProfile(),
				BigDecimal.valueOf(utilityGoal),
				BigDecimal.valueOf(utilityGoal + tolerance));
		// pick a random one.
		return options.get(new Random().nextInt(options.size().intValue()));

	}

	/**
	 * 
	 * @param t the time in [0,1] where 0 means start of nego and 1 the end of
	 *          nego (absolute time/round limit)
	 * @param e the e value
	 * @return the utility goal for this time and e value
	 */
	private double utilityGoal(double t, double e) {
		double ft = 0;
		if (e != 0)
			ft = Math.pow(t, 1 / e);
		// we subtract 0.000001 to correct possibly small round-up errors
		return minUtil + (maxUtil - minUtil) * (1 - ft) - 0.000001;
	}
}
